# Add relast2uml to your project

If you want to use `Relast2Uml`, either use the latest [pre-build version](#use-packaged-version) or clone the repository and [build it yourself](#build-from-source).

## Use packaged version

Check the [package overview page](https://git-st.inf.tu-dresden.de/jastadd/relast2uml/-/packages) to find the latest versions of the individual packages.

First add this GitLab as a repository in your `build.gradle`:

```
repositories {
    maven {
        name "gitlab-maven"
        url "https://git-st.inf.tu-dresden.de/api/v4/groups/jastadd/-/packages/maven"
    }
}
```

### grammar2uml

To use `Grammar2uml`, add it as a dependency:

```
configurations {
    grammar2umlClasspath
}

dependencies {
    grammar2umlClasspath group: 'de.tudresden.inf.st', name: 'grammar2uml', version: '0.1.1'
}
```

Finally, add a task to create your visualization:

```
task grammar2uml(type: JavaExec) {
    main = 'de.tudresden.inf.st.jastadd.grammar2uml.compiler.Compiler'
    classpath = configurations.grammar2umlClasspath

    args([
            '--verbose',
            'src/main/jastadd/GoalModel.relast'
    ])
}
```

### dumpAst

To use `DumpAst` or `DumpAstWithPlantuml`, add it as a dependency:

```
dependencies {
    implementation group: 'de.tudresden.inf.st', name: 'dumpAst', version: '0.3.4'
    // or
    implementation group: 'de.tudresden.inf.st', name: 'dumpAstWithPlantuml', version: '0.3.4'
}
```


## Build from source

If you want to build the tools of `Relast2Uml` from source, first build the jar from the [Relast2Uml repository](https://git-st.inf.tu-dresden.de/jastadd/relast2uml):

```bash
git clone https://git-st.inf.tu-dresden.de/jastadd/relast2uml.git
cd relast2uml
./gradlew jar
ls grammar2uml/build/libs/ dumpAst/build/libs/ dumpAstWithPlantuml/build/libs/
```

Those JARs can then be copied to your project, e.g. for grammar2uml.

```bash
cp grammar2uml/build/libs/grammar2uml-<version>.jar ../your-project/libs/grammar2uml.jar
cd ../your-project/
```

Finally, this JAR has to be integrated into your build process. In case, [Gradle](https://gradle.org/) is used, the JAR file needs to be added as dependency using:

```groovy
dependencies {
    implementation fileTree(include: ['grammar2uml.jar'], dir: 'libs')
    implementation fileTree(include: ['dumpAst.jar'], dir: 'libs')
}
```

The path to the JAR file may need to be changed according to your project structure.
In case of `grammar2uml` a task needs to be created, similar to the one defined [above](#grammar2uml).

```groovy
task grammar2uml(type: JavaExec) {
    main = '-jar'

    args([
            '../libs/grammar2uml.jar',
            '--verbose',
            'src/main/jastadd/GoalModel.relast'
    ])
}
```
