package de.tudresden.inf.st.jastadd.testDumper;

import de.tudresden.inf.st.jastadd.dumpAst.ast.DumpNode;
import org.jastadd.testDumper.ast.Root;
import org.junit.jupiter.api.Test;

import java.util.List;

import static de.tudresden.inf.st.jastadd.testDumper.TestUtils.*;
import static org.assertj.core.api.Assertions.*;

public class TestIncluded {

  @Test
  public void testValueAttributeDefault() {
    Root root = createRoot(null, null);

    List<DumpNode> nodes = TestUtils.dumpModel(root);
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactly(ROOT_NAME);
    DumpNode actualRoot = TestUtils.findByName(nodes, ROOT_NAME);
    assertThat(valueTokens(actualRoot)).containsOnly(entry("Name", ROOT_NAME));
  }

  @Test
  public void testValueAttributeIncluded() {
    Root root = createRoot(null, null);

    List<DumpNode> nodes = TestUtils.dumpModel(root, db -> db.includeAttributes("simpleAttr"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactly(ROOT_NAME);
    DumpNode actualRoot = TestUtils.findByName(nodes, ROOT_NAME);
    assertThat(valueTokens(actualRoot)).containsOnly(entry("Name", ROOT_NAME), entry("simpleAttr", 42));
  }

  @Test
  public void testReferenceAttributeDefault() {
    Root root = createRoot(createA(A_NAME), null);

    List<DumpNode> nodes = TestUtils.dumpModel(root);
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactly(ROOT_NAME, A_NAME);
    DumpNode actualRoot = TestUtils.findByName(nodes, ROOT_NAME);
    assertThat(referenceTokens(actualRoot)).isEmpty();
  }

    @Test
  public void testReferenceAttributeIncluded() {
    Root root = createRoot(createA(A_NAME), null);

    List<DumpNode> nodes = TestUtils.dumpModel(root, db -> db.includeAttributes("referenceAttr"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactly(ROOT_NAME, A_NAME);
    DumpNode actualRoot = TestUtils.findByName(nodes, ROOT_NAME);
    assertThatMapOf(referenceTokens(actualRoot)).containsOnly(tuple("referenceAttr", A_NAME));
  }

  @Test
  public void testNormalNTADefault() {
    Root root = createRoot(null, createC(C_NAME));

    List<DumpNode> nodes = TestUtils.dumpModel(root);
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactly(ROOT_NAME, C_NAME);
    DumpNode actualC = TestUtils.findByName(nodes, C_NAME);
    assertThatMapOf(normalChildren(actualC)).isEmpty();
  }

  @Test
  public void testNormalNTAIncluded() {
    Root root = createRoot(null, createC(C_NAME));

    List<DumpNode> nodes = TestUtils.dumpModel(root, db -> db.includeNonterminalAttributes("Calculated"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactly(ROOT_NAME, C_NAME, "Calculated-" + C_NAME);
    DumpNode actualC = TestUtils.findByName(nodes, C_NAME);
    assertThatMapOf(normalChildren(actualC)).containsOnly(tuple("Calculated", "Calculated-" + C_NAME));
  }

  @Test
  public void testListNTADefault() {
    Root root = createRoot(null, createC(C_NAME));

    List<DumpNode> nodes = TestUtils.dumpModel(root);
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactly(ROOT_NAME, C_NAME);
    DumpNode actualC = TestUtils.findByName(nodes, C_NAME);
    assertThat(listChildren(actualC)).isEmpty();

  }

  @Test
  public void testListNTAIncluded() {
    Root root = createRoot(null, createC(C_NAME));

    List<DumpNode> nodes = TestUtils.dumpModel(root, db -> db.includeNonterminalAttributes("AlsoCalculated"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactly(ROOT_NAME, C_NAME, "AlsoCalculated-" + C_NAME);
    DumpNode actualC = TestUtils.findByName(nodes, C_NAME);
    assertThatMapOf(listChildren(actualC), "AlsoCalculated").containsExactly("AlsoCalculated-" + C_NAME);
  }
}
