package de.tudresden.inf.st.jastadd.testDumper;

import de.tudresden.inf.st.jastadd.dumpAst.ast.DumpNode;
import org.jastadd.testDumper.ast.A;
import org.jastadd.testDumper.ast.AbstractT;
import org.jastadd.testDumper.ast.TRoot;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;

import java.util.List;
import java.util.function.Consumer;

import static de.tudresden.inf.st.jastadd.testDumper.TestUtils.*;
import static org.assertj.core.api.Assertions.*;

/**
 * Testing type-level exclusions.
 * <p>
 * Refer to {@link de.tudresden.inf.st.jastadd.dumpAst.ast.DumpBuilder DumpBuilder} for levels of inclusion/exclusion.
 *
 * @author rschoene - Initial contribution
 */
public class TestTypeLevel3 {

  @Test
  public void testTokenLevel2Excluded() {
    Consumer<AbstractT> setUnwanted = t -> t.setUnwanted(5);
    TRoot root = createTRoot(createA(A_NAME), createT1(setUnwanted), createT2(setUnwanted), createT3(setUnwanted));
    Assertions.assertEquals(5, root.getT2().getUnwanted());

    List<DumpNode> nodes = TestUtils.dumpModel(root, dp -> dp.excludeTokens("Unwanted"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME);
    for (String name : Tx_NAMES) {
      DumpNode actualTx = TestUtils.findByName(nodes, name);
      assertThat(valueTokens(actualTx)).containsOnly(entry("Name", name));
    }
  }

  @Test
  public void testTokenLevel3SomeIncluded() {
    Consumer<AbstractT> setUnwanted = t -> t.setUnwanted(5);
    TRoot root = createTRoot(createA(A_NAME), createT1(setUnwanted), createT2(setUnwanted), createT3(setUnwanted));
    Assertions.assertEquals(5, root.getT2().getUnwanted());

    List<DumpNode> nodesT2 = TestUtils.dumpModel(root,
        dp -> dp.excludeTokens("Unwanted")
            .includeTokensFor("T2", "Unwanted"));
    assertThat(nodesT2).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME);
    assertThat(valueTokens(TestUtils.findByName(nodesT2, T1_NAME))).containsOnly(
        entry("Name", T1_NAME));
    assertThat(valueTokens(TestUtils.findByName(nodesT2, T2_NAME))).containsOnly(
        entry("Name", T2_NAME), entry("Unwanted", 5));
    assertThat(valueTokens(TestUtils.findByName(nodesT2, T3_NAME))).containsOnly(
        entry("Name", T3_NAME));

    List<DumpNode> nodesT2T3 = TestUtils.dumpModel(root,
        dp -> dp.excludeTokens("Unwanted")
            .includeTokensFor("T2|T3", "Unwanted"));
    assertThat(nodesT2T3).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME);
    assertThat(valueTokens(TestUtils.findByName(nodesT2T3, T1_NAME))).containsOnly(
        entry("Name", T1_NAME));
    assertThat(valueTokens(TestUtils.findByName(nodesT2T3, T2_NAME))).containsOnly(
        entry("Name", T2_NAME), entry("Unwanted", 5));
    assertThat(valueTokens(TestUtils.findByName(nodesT2T3, T3_NAME))).containsOnly(
        entry("Name", T3_NAME), entry("Unwanted", 5));
  }

  @Test
  public void testNormalChildLevel2Excluded() {
    Consumer<AbstractT> setB = t -> t.setB(createB(t.getName() + B_NAME));
    TRoot root = createTRoot(createA(A_NAME), createT1(setB), createT2(setB), createT3(setB));
    Assertions.assertEquals(T2_NAME + B_NAME, root.getT2().getB().getName());

    List<DumpNode> nodes = TestUtils.dumpModel(root, dp -> dp.excludeChildren("B"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME);
    for (String name : Tx_NAMES) {
      DumpNode actualTx = TestUtils.findByName(nodes, name);
      assertThat(normalChildren(actualTx)).isEmpty();
    }
  }

  @Test
  public void testNormalChildLevel3SomeIncluded() {
    Consumer<AbstractT> setB = t -> t.setB(createB(t.getName() + B_NAME));
    TRoot root = createTRoot(createA(A_NAME), createT1(setB), createT2(setB), createT3(setB));
    Assertions.assertEquals(T2_NAME + B_NAME, root.getT2().getB().getName());

    List<DumpNode> nodes = TestUtils.dumpModel(root,
        dp -> dp.excludeChildren("B")
            .includeChildrenFor("T2", "B"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME, T2_NAME + B_NAME);
    assertThat(normalChildren(TestUtils.findByName(nodes, T1_NAME))).isEmpty();
    assertThatMapOf(normalChildren(TestUtils.findByName(nodes, T2_NAME))).containsOnly(
        tuple("B", T2_NAME + B_NAME));
    assertThat(normalChildren(TestUtils.findByName(nodes, T3_NAME))).isEmpty();
  }

  @Test
  public void testListChildLevel2Excluded() {
    Consumer<AbstractT> addBee = t -> t.addBee(createB(t.getName() + B_NAME));
    TRoot root = createTRoot(createA(A_NAME), createT1(addBee), createT2(addBee), createT3(addBee));
    Assertions.assertEquals(T2_NAME + B_NAME, root.getT2().getBee(0).getName());

    List<DumpNode> nodes = TestUtils.dumpModel(root,
        dp -> dp.excludeChildren("Bee"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME);
    for (String name : Tx_NAMES) {
      assertThat(listChildren(TestUtils.findByName(nodes, name))).isEmpty();
    }
  }

  @Test
  public void testListChildLevel3SomeIncluded() {
    Consumer<AbstractT> addBee = t -> t.addBee(createB(t.getName() + B_NAME));
    TRoot root = createTRoot(createA(A_NAME), createT1(addBee), createT2(addBee), createT3(addBee));
    Assertions.assertEquals(T2_NAME + B_NAME, root.getT2().getBee(0).getName());

    List<DumpNode> nodes = TestUtils.dumpModel(root,
        dp -> dp.excludeChildren("Bee")
                .includeChildrenFor("T2", "Bee"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME, T2_NAME + B_NAME);
    assertThat(listChildren(TestUtils.findByName(nodes, T1_NAME))).isEmpty();
    assertThatMapOf(listChildren(TestUtils.findByName(nodes, T2_NAME)), "Bee").containsOnly(
        T2_NAME + B_NAME);
    assertThat(listChildren(TestUtils.findByName(nodes, T3_NAME))).isEmpty();
  }

  @Test
  public void testRelationLevel2Excluded() {
    final A a = createA(A_NAME);
    Consumer<AbstractT> setOneA = t -> t.setOneA(a);
    TRoot root = createTRoot(a, createT1(setOneA), createT2(setOneA), createT3(setOneA));
    Assertions.assertEquals(a, root.getT2().getOneA());

    List<DumpNode> nodes = TestUtils.dumpModel(root, dp -> dp.excludeRelations("OneA"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME);
    for (String name : Tx_NAMES) {
      DumpNode actualTx = TestUtils.findByName(nodes, name);
      assertThat(normalRelationChildren(actualTx)).isEmpty();
    }
  }

  @Test
  public void testRelationLevel3SomeIncluded() {
    final A a = createA(A_NAME);
    Consumer<AbstractT> setOneA = t -> t.setOneA(a);
    TRoot root = createTRoot(a, createT1(setOneA), createT2(setOneA), createT3(setOneA));
    Assertions.assertEquals(a, root.getT2().getOneA());

    List<DumpNode> nodes = TestUtils.dumpModel(root,
        dp -> dp.excludeRelations("OneA")
            .includeRelationsFor("T2", "OneA"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME);
    assertThat(normalRelationChildren(TestUtils.findByName(nodes, T1_NAME))).isEmpty();
    assertThatMapOf(normalRelationChildren(TestUtils.findByName(nodes, T2_NAME))).containsOnly(
        tuple("OneA", A_NAME));
    assertThat(normalRelationChildren(TestUtils.findByName(nodes, T3_NAME))).isEmpty();
  }

  @Test
  public void testAttributeLevel2Included() {
    TRoot root = createTRoot(createA(A_NAME), createT1(), createT2(), createT3());

    List<DumpNode> nodes = dumpModel(root,
        dp -> dp.includeAttributes("simpleAttr"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME);
    for (String name : Tx_NAMES) {
      DumpNode actualTx = findByName(nodes, name);
      assertThat(valueTokens(actualTx)).containsOnly(
          entry("Name", name),
          entry(TOKEN_LABEL_UNWANTED, 0),
          entry("simpleAttr", 43));
    }
  }

  @Test
  public void testAttributeLevel3SomeExcluded() {
    TRoot root = createTRoot(createA(A_NAME), createT1(), createT2(), createT3());

    List<DumpNode> nodes = dumpModel(root,
        dp -> dp.includeAttributes("simpleAttr")
                .excludeAttributesFor("T2", "simpleAttr"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME);
    assertThat(valueTokens(findByName(nodes, T1_NAME))).containsOnly(
        entry("Name", T1_NAME), entry(TOKEN_LABEL_UNWANTED, 0), entry("simpleAttr", 43));
    assertThat(valueTokens(findByName(nodes, T2_NAME))).containsOnly(
        entry("Name", T2_NAME), entry(TOKEN_LABEL_UNWANTED, 0));
    assertThat(valueTokens(findByName(nodes, T3_NAME))).containsOnly(
        entry("Name", T3_NAME), entry(TOKEN_LABEL_UNWANTED, 0), entry("simpleAttr", 43));
  }

  @Test
  public void testNTALevel2Included() {
    TRoot root = createTRoot(createA(A_NAME), createT1(), createT2(), createT3());

    List<DumpNode> nodes = dumpModel(root,
        dp -> dp.includeNonterminalAttributes("Calculated"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME, "Calculated-" + T1_NAME, "Calculated-" + T2_NAME, "Calculated-" + T3_NAME);
    for (String name : Tx_NAMES) {
      assertThatMapOf(normalChildren(findByName(nodes, name))).containsOnly(
          tuple("Calculated", "Calculated-" + name));
    }
  }

  @Test
  public void testNTALevel3SomeExcluded() {
    TRoot root = createTRoot(createA(A_NAME), createT1(), createT2(), createT3());

    List<DumpNode> nodes = dumpModel(root,
        dp -> dp.includeNonterminalAttributes("Calculated")
            .excludeNonterminalAttributesFor("T2", "Calculated"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME, "Calculated-" + T1_NAME, "Calculated-" + T3_NAME);
    assertThatMapOf(normalChildren(findByName(nodes, T1_NAME))).containsOnly(
        tuple("Calculated", "Calculated-" + T1_NAME));
    assertThat(normalChildren(findByName(nodes, T2_NAME))).isEmpty();
    assertThatMapOf(normalChildren(findByName(nodes, T3_NAME))).containsOnly(
        tuple("Calculated", "Calculated-" + T3_NAME));
  }

}
