package de.tudresden.inf.st.jastadd.testDumper;

import de.tudresden.inf.st.jastadd.dumpAst.ast.DumpNode;
import org.jastadd.testDumper.ast.A;
import org.jastadd.testDumper.ast.AbstractT;
import org.jastadd.testDumper.ast.TRoot;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.Test;

import java.util.List;
import java.util.function.Consumer;

import static de.tudresden.inf.st.jastadd.testDumper.TestUtils.*;
import static org.assertj.core.api.Assertions.*;

/**
 * Testing type-level inclusions.
 * <p>
 * Refer to {@link de.tudresden.inf.st.jastadd.dumpAst.ast.DumpBuilder DumpBuilder} for levels of inclusion/exclusion.
 *
 * @author rschoene - Initial contribution
 */
public class TestTypeLevel4 {

  @Test
  public void testTokenLevel3Included() {
    Consumer<AbstractT> setUnwanted = t -> t.setUnwanted(5);
    TRoot root = createTRoot(createA(A_NAME, createC(C_NAME, c -> c.setUnwanted(6))),
        createT1(setUnwanted), createT2(setUnwanted), createT3(setUnwanted));
    Assertions.assertEquals(5, root.getT2().getUnwanted());

    List<DumpNode> nodes = dumpModel(root,
        dp -> dp.excludeTokens(TOKEN_LABEL_UNWANTED)
            .includeTokensFor("T.", TOKEN_LABEL_UNWANTED));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(
        ROOT_NAME, A_NAME, C_NAME, T1_NAME, T2_NAME, T3_NAME);
    for (String name : Tx_NAMES) {
      DumpNode actualTx = findByName(nodes, name);
      assertThat(valueTokens(actualTx)).containsOnly(entry("Name", name), entry(TOKEN_LABEL_UNWANTED, 5));
    }
    assertThat(valueTokens(findByName(nodes, C_NAME))).containsOnly(
        entry("Name", C_NAME));
  }

  @Test
  public void testTokenLevel4SomeExcluded() {
    Consumer<AbstractT> setUnwanted = t -> t.setUnwanted(5);
    TRoot root = createTRoot(createA(A_NAME, createC(C_NAME, c -> c.setUnwanted(6))),
        createT1(setUnwanted), createT2(setUnwanted), createT3(setUnwanted));
    Assertions.assertEquals(5, root.getT2().getUnwanted());

    List<DumpNode> nodesT2 = dumpModel(root,
        dp -> dp.excludeTokens(TOKEN_LABEL_UNWANTED)
            .includeTokensFor("T.", TOKEN_LABEL_UNWANTED)
            .excludeTokensFor("T3", TOKEN_LABEL_UNWANTED));
    assertThat(nodesT2).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(
        ROOT_NAME, A_NAME, C_NAME, T1_NAME, T2_NAME, T3_NAME);
    assertThat(valueTokens(findByName(nodesT2, T1_NAME))).containsOnly(
        entry("Name", T1_NAME), entry(TOKEN_LABEL_UNWANTED, 5));
    assertThat(valueTokens(findByName(nodesT2, T2_NAME))).containsOnly(
        entry("Name", T2_NAME), entry(TOKEN_LABEL_UNWANTED, 5));
    assertThat(valueTokens(findByName(nodesT2, T3_NAME))).containsOnly(
        entry("Name", T3_NAME));
    assertThat(valueTokens(findByName(nodesT2, C_NAME))).containsOnly(
        entry("Name", C_NAME));

    List<DumpNode> nodesT2T3 = dumpModel(root,
        dp -> dp.excludeTokens(TOKEN_LABEL_UNWANTED)
            .includeTokensFor("T.", TOKEN_LABEL_UNWANTED)
            .excludeTokensFor("T2|T3", TOKEN_LABEL_UNWANTED));
    assertThat(nodesT2T3).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(
        ROOT_NAME, A_NAME, C_NAME, T1_NAME, T2_NAME, T3_NAME);
    assertThat(valueTokens(findByName(nodesT2T3, T1_NAME))).containsOnly(
        entry("Name", T1_NAME), entry(TOKEN_LABEL_UNWANTED, 5));
    assertThat(valueTokens(findByName(nodesT2T3, T2_NAME))).containsOnly(
        entry("Name", T2_NAME));
    assertThat(valueTokens(findByName(nodesT2T3, T3_NAME))).containsOnly(
        entry("Name", T3_NAME));
    assertThat(valueTokens(findByName(nodesT2T3, C_NAME))).containsOnly(
        entry("Name", C_NAME));
  }

  @Test
  public void testNormalChildLevel3Included() {
    Consumer<AbstractT> setB = t -> t.setB(createB(t.getName() + B_NAME));
    TRoot root = createTRoot(createA(A_NAME), createT1(setB), createT2(setB), createT3(setB));
    Assertions.assertEquals(T2_NAME + B_NAME, root.getT2().getB().getName());

    List<DumpNode> nodes = dumpModel(root,
        dp -> dp.excludeChildren("B")
            .includeChildrenFor("T.", "B"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME, T1_NAME + B_NAME, T2_NAME + B_NAME, T3_NAME + B_NAME);
    for (String name : Tx_NAMES) {
      assertThatMapOf(normalChildren(findByName(nodes, name))).containsOnly(
          tuple("B", name + B_NAME));
    }
  }

  @Test
  public void testNormalChildLevel4SomeExcluded() {
    Consumer<AbstractT> setB = t -> t.setB(createB(t.getName() + B_NAME));
    TRoot root = createTRoot(createA(A_NAME), createT1(setB), createT2(setB), createT3(setB));
    Assertions.assertEquals(T2_NAME + B_NAME, root.getT2().getB().getName());

    List<DumpNode> nodes = dumpModel(root,
        dp -> dp.excludeChildren("B")
            .includeChildrenFor("T.", "B")
            .excludeChildrenFor("T3", "B"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME, T1_NAME + B_NAME, T2_NAME + B_NAME);
    assertThatMapOf(normalChildren(findByName(nodes, T1_NAME))).containsOnly(
        tuple("B", T1_NAME + B_NAME));
    assertThatMapOf(normalChildren(findByName(nodes, T2_NAME))).containsOnly(
        tuple("B", T2_NAME + B_NAME));
    assertThat(normalChildren(findByName(nodes, T3_NAME))).isEmpty();
  }

  @Test
  public void testListChildLevel3Included() {
    Consumer<AbstractT> addBee = t -> t.addBee(createB(t.getName() + B_NAME));
    TRoot root = createTRoot(createA(A_NAME), createT1(addBee), createT2(addBee), createT3(addBee));
    Assertions.assertEquals(T3_NAME + B_NAME, root.getT3().getBee(0).getName());

    List<DumpNode> nodes = dumpModel(root,
        dp -> dp.excludeChildren("Bee")
            .includeChildrenFor("T.", "Bee"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME, T1_NAME + B_NAME, T2_NAME + B_NAME, T3_NAME + B_NAME);
    for (String name : Tx_NAMES) {
      assertThatMapOf(listChildren(findByName(nodes, name)), "Bee").containsOnly(
          name + B_NAME);
    }
  }

  @Test
  public void testListChildLevel4SomeExcluded() {
    Consumer<AbstractT> addBee = t -> t.addBee(createB(t.getName() + B_NAME));
    TRoot root = createTRoot(createA(A_NAME), createT1(addBee), createT2(addBee), createT3(addBee));
    Assertions.assertEquals(T3_NAME + B_NAME, root.getT3().getBee(0).getName());

    List<DumpNode> nodes = dumpModel(root,
        dp -> dp.excludeChildren("Bee")
            .includeChildrenFor("T.", "Bee")
            .excludeChildrenFor("T3", "Bee"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME, T1_NAME + B_NAME, T2_NAME + B_NAME);
    assertThatMapOf(listChildren(findByName(nodes, T1_NAME)), "Bee").containsOnly(
        T1_NAME + B_NAME);
    assertThatMapOf(listChildren(findByName(nodes, T2_NAME)), "Bee").containsOnly(
        T2_NAME + B_NAME);
    assertThat(listChildren(findByName(nodes, T3_NAME))).isEmpty();
  }

  @Test
  public void testRelationLevel3Included() {
    final A a = createA(A_NAME);
    Consumer<AbstractT> setOneA = t -> t.setOneA(a);
    TRoot root = createTRoot(a, createT1(setOneA), createT2(setOneA), createT3(setOneA));
    Assertions.assertEquals(a, root.getT2().getOneA());

    List<DumpNode> nodes = dumpModel(root,
        dp -> dp.excludeRelations("OneA")
            .includeRelationsFor("T2", "OneA"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME);
    assertThat(normalRelationChildren(findByName(nodes, T1_NAME))).isEmpty();
    assertThatMapOf(normalRelationChildren(findByName(nodes, T2_NAME))).containsOnly(
        tuple("OneA", A_NAME));
    assertThat(normalRelationChildren(findByName(nodes, T3_NAME))).isEmpty();
  }

  @Test
  public void testRelationLevel4SomeExcluded() {
    final A a = createA(A_NAME);
    Consumer<AbstractT> setOneA = t -> t.setOneA(a);
    TRoot root = createTRoot(a, createT1(setOneA), createT2(setOneA), createT3(setOneA));
    Assertions.assertEquals(a, root.getT2().getOneA());

    List<DumpNode> nodes = dumpModel(root,
        dp -> dp.excludeRelations("OneA")
            .includeRelationsFor("T.", "OneA")
            .excludeRelationsFor("T3", "OneA"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME);
    assertThatMapOf(normalRelationChildren(findByName(nodes, T1_NAME))).containsOnly(
        tuple("OneA", A_NAME));
    assertThatMapOf(normalRelationChildren(findByName(nodes, T2_NAME))).containsOnly(
        tuple("OneA", A_NAME));
    assertThat(normalRelationChildren(findByName(nodes, T3_NAME))).isEmpty();
  }

  @Test
  public void testAttributeLevel3Excluded() {
    TRoot root = createTRoot(createA(A_NAME), createT1(), createT2(), createT3());

    List<DumpNode> nodes = TestUtils.dumpModel(root,
        dp -> dp.includeAttributes("simpleAttr")
                .excludeAttributesFor("T.", "simpleAttr"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME);
    for (String name : Tx_NAMES) {
      DumpNode actualTx = TestUtils.findByName(nodes, name);
      assertThat(valueTokens(actualTx)).containsOnly(
          entry("Name", name), entry(TOKEN_LABEL_UNWANTED, 0));
    }
  }

  @Test
  public void testAttributeLevel4SomeIncluded() {
    TRoot root = createTRoot(createA(A_NAME), createT1(), createT2(), createT3());

    List<DumpNode> nodes = TestUtils.dumpModel(root,
        dp -> dp.includeAttributes("simpleAttr")
                .excludeAttributesFor("T.", "simpleAttr")
                .includeAttributesFor("T3", "simpleAttr"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME);
    assertThat(valueTokens(TestUtils.findByName(nodes, T1_NAME))).containsOnly(
        entry("Name", T1_NAME), entry(TOKEN_LABEL_UNWANTED, 0));
    assertThat(valueTokens(TestUtils.findByName(nodes, T2_NAME))).containsOnly(
        entry("Name", T2_NAME), entry(TOKEN_LABEL_UNWANTED, 0));
    assertThat(valueTokens(TestUtils.findByName(nodes, T3_NAME))).containsOnly(
        entry("Name", T3_NAME), entry(TOKEN_LABEL_UNWANTED, 0), entry("simpleAttr", 43));
  }

  @Test
  public void testNTALevel3Excluded() {
    TRoot root = createTRoot(createA(A_NAME), createT1(), createT2(), createT3());

    List<DumpNode> nodes = TestUtils.dumpModel(root,
        dp -> dp.includeNonterminalAttributes("Calculated")
                .excludeNonterminalAttributesFor("T.", "Calculated"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME);
    DumpNode actualT3 = TestUtils.findByName(nodes, T3_NAME);
    assertThat(normalChildren(actualT3)).isEmpty();
  }

  @Test
  public void testNTALevel4SomeIncluded() {
    TRoot root = createTRoot(createA(A_NAME), createT1(), createT2(), createT3());

    List<DumpNode> nodes = TestUtils.dumpModel(root,
        dp -> dp.includeNonterminalAttributes("Calculated")
            .excludeNonterminalAttributesFor("T.", "Calculated")
            .includeNonterminalAttributesFor("T3", "Calculated"));
    assertThat(nodes).flatExtracting(NAME_EXTRACTOR).containsExactlyInAnyOrder(ROOT_NAME, A_NAME, T1_NAME, T2_NAME, T3_NAME, "Calculated-" + T3_NAME);
    DumpNode actualT3 = TestUtils.findByName(nodes, T3_NAME);
    assertThatMapOf(normalChildren(actualT3)).containsOnly(tuple("Calculated", "Calculated-" + T3_NAME));
  }

}
