package de.tudresden.inf.st.jastadd.testDumper;

import de.tudresden.inf.st.jastadd.dumpAst.ast.*;
import org.assertj.core.api.AbstractListAssert;
import org.assertj.core.api.Assertions;
import org.assertj.core.api.ObjectAssert;
import org.assertj.core.groups.Tuple;
import org.jastadd.testDumper.ast.*;

import java.util.*;
import java.util.function.Consumer;
import java.util.function.Function;

import static org.junit.jupiter.api.Assertions.assertFalse;

public class TestUtils {
  public static final Random rand = new Random();

  public static final String TOKEN_LABEL_OTHER_VALUE = "OtherValue";
  public static final String TOKEN_LABEL_UNWANTED = "Unwanted";
  public static final String TOKEN_LABEL_RAW_REFERENCE = "RawReference";

  public static final String ROOT_NAME = "Root" + Integer.toHexString(rand.nextInt(0xFFFFFF));
  public static final String A_NAME = "A" + Integer.toHexString(rand.nextInt(0xFFFFFF));
  public static final String B_NAME = "B" + Integer.toHexString(rand.nextInt(0xFFFFFF));
  public static final String B1_NAME = "B1" + Integer.toHexString(rand.nextInt(0xFFFFFF));
  public static final String B2_NAME = "B2" + Integer.toHexString(rand.nextInt(0xFFFFFF));
  public static final String B3_NAME = "B3" + Integer.toHexString(rand.nextInt(0xFFFFFF));
  public static final String C_NAME = "C" + Integer.toHexString(rand.nextInt(0xFFFFFF));
  public static final String T1_NAME = "T1" + Integer.toHexString(rand.nextInt(0xFFFFFF));
  public static final String T2_NAME = "T2" + Integer.toHexString(rand.nextInt(0xFFFFFF));
  public static final String T3_NAME = "T3" + Integer.toHexString(rand.nextInt(0xFFFFFF));
  public static final String[] Tx_NAMES = {T1_NAME, T2_NAME, T3_NAME};

  public static Root createRoot(A a, C c, B... bs) {
    Root result = new Root();
    result.setName(ROOT_NAME);
    if (a != null) {
      result.setA(a);
    }
    if (c != null) {
      result.setC(c);
    }
    for (B b : bs) {
      result.addB(b);
    }
    return result;
  }

  public static A createA(String name) {
    return createA(name, a -> {});
  }

  public static A createA(String name, Consumer<A> additionalSettings) {
    A result = new A();
    result.setName(name);
    additionalSettings.accept(result);
    return result;
  }

  public static A createA(String name, C myC) {
    return createA(name, a -> a.setMyC(myC));
  }

  public static B createB(String name) {
    return createB(name, b -> {});
  }

  public static B createB(String name, String otherValue) {
    return createB(name, b -> b.setOtherValue(otherValue));
  }

  public static B createB(String name, Consumer<B> additionalSettings) {
    B result = new B();
    result.setName(name);
    additionalSettings.accept(result);
    return result;
  }

  public static C createC(String name) {
    return createC(name, c -> {});
  }

  public static C createC(String name, A optA) {
    return createC(name, c -> c.setA(optA));
  }

  public static C createC(String name, Consumer<C> additionalSettings) {
    C result = new C();
    result.setName(name);
    additionalSettings.accept(result);
    return result;
  }

  public static TRoot createTRoot(A a, T1 t1, T2 t2, T3 t3) {
    TRoot result = new TRoot();
    result.setName(ROOT_NAME);
    if (a != null) {
      result.setA(a);
    }
    if (t1 != null) {
      result.setT1(t1);
    }
    if (t2 != null) {
      result.setT2(t2);
    }
    if (t3 != null) {
      result.setT3(t3);
    }
    return result;
  }

  @SafeVarargs
  public static T1 createT1(Consumer<AbstractT>... additionalSettings) {
    return setupAbstractT(new T1(), T1_NAME, additionalSettings);
  }

  @SafeVarargs
  public static T2 createT2(Consumer<AbstractT>... additionalSettings) {
    return setupAbstractT(new T2(), T2_NAME, additionalSettings);
  }

  @SafeVarargs
  public static T3 createT3(Consumer<AbstractT>... additionalSettings) {
    return setupAbstractT(new T3(), T3_NAME, additionalSettings);
  }

  private static <T extends AbstractT> T setupAbstractT(T t, String name, Consumer<AbstractT>[] additionalSettings) {
    t.setName(name);
    for (Consumer<AbstractT> setting : additionalSettings) {
      setting.accept(t);
    }
    return t;
  }

  public static final Function<DumpNode, String> NAME_EXTRACTOR = node -> {
    for (DumpToken dumpToken : node.getDumpTokenList()) {
      if (dumpToken.getName().equals("Name")) {
        return dumpToken.asDumpValueToken().getValue().toString();
      }
    }
    return null;
  };

  public static AbstractListAssert<?, List<? extends Tuple>, Tuple, ObjectAssert<Tuple>> assertThatMapOf(Map<String, DumpNode> map) {
    return Assertions.assertThat(map).extractingFromEntries(Map.Entry::getKey, e -> NAME_EXTRACTOR.apply(e.getValue()));
  }

  public static AbstractListAssert<?, List<?>, Object, ObjectAssert<Object>> assertThatMapOf(Map<String, List<DumpNode>> map, String key) {
    return Assertions.assertThat(map.get(key)).flatExtracting(NAME_EXTRACTOR);
  }

  public static List<DumpNode> dumpModel(Object target) {
    return dumpModel(target, db -> {});
  }

  public static List<DumpNode> dumpModel(Object target, Consumer<DumpBuilder> options) {
    ExposingDumpBuilder builder = new ExposingDumpBuilder(target);
    options.accept(builder);
    DumpAst dumpAst = builder.build();
    // let mustache run, but ignore result
    dumpAst.toPlantUml();
    List<DumpNode> result = new ArrayList<>();
    for (DumpNode dumpNode : dumpAst.getDumpNodeList()) {
      if (dumpNode.isAstNode() && !dumpNode.getInvisible()) {
        result.add(dumpNode);
      }
    }
    return result;
  }

  public static DumpNode findByName(List<DumpNode> nodes, String name) {
    for (DumpNode node : nodes) {
      if (name.equals(NAME_EXTRACTOR.apply(node))) {
        return node;
      }
    }
    Assertions.fail("No node found with name " + name);
    // DummyCode to keep intellij silent
    return new DumpNode();
  }

  public static Map<String, DumpNode> normalChildren(DumpNode node) {
    Map<String, DumpNode> result = new HashMap<>();
    for (DumpChildNode dumpChildNode : node.getDumpChildNodeList()) {
      if (!dumpChildNode.isList()) {
        // then it is a DumpNormalChildNode
        DumpNode target = ((DumpNormalChildNode) dumpChildNode).getDumpNode();
        if (!target.getInvisible()) {
          result.put(dumpChildNode.getName(), target);
        }
      }
    }
    return result;
  }

  public static Map<String, List<DumpNode>> listChildren(DumpNode node) {
    Map<String, List<DumpNode>> result = new HashMap<>();
    for (DumpChildNode dumpChildNode : node.getDumpChildNodeList()) {
      if (dumpChildNode.isList()) {
        // then it is a DumpListChildNode
        ((DumpListChildNode) dumpChildNode).getInnerDumpNodeList().forEach(inner -> {
          if (!inner.getDumpNode().getInvisible()) {
            result.computeIfAbsent(dumpChildNode.getName(), key -> new ArrayList<>()).add(inner.getDumpNode());
          }
        });
      }
    }
    return result;
  }

  public static Map<String, DumpNode> normalRelationChildren(DumpNode node) {
    Map<String, DumpNode> result = new HashMap<>();
    for (DumpRelation dumpRelation : node.getDumpRelationList()) {
      if (!dumpRelation.isList()) {
        // then it is a DumpNormalRelation
        DumpNode target = ((DumpNormalRelation) dumpRelation).getDumpNode();
        if (!target.getInvisible()) {
          result.put(dumpRelation.getName(), target);
        }
      }
    }
    return result;
  }

  public static Map<String, List<DumpNode>> listRelationChildren(DumpNode node) {
    Map<String, List<DumpNode>> result = new HashMap<>();
    for (DumpRelation dumpRelation : node.getDumpRelationList()) {
      if (dumpRelation.isList()) {
        // then it is a DumpListRelation
        ((DumpListRelation) dumpRelation).getInnerDumpNodeList().forEach(inner -> {
          if (!inner.getDumpNode().getInvisible()) {
            result.computeIfAbsent(dumpRelation.getName(), key -> new ArrayList<>()).add(inner.getDumpNode());
          }
        });
      }
    }
    return result;
  }

  public static Map<String, DumpNode> referenceTokens(DumpNode node) {
    Map<String, DumpNode> result = new HashMap<>();
    for (DumpToken dumpToken : node.getDumpTokenList()) {
      if (!dumpToken.isDumpValueToken()) {
        // then it is a DumpReferenceToken
        DumpNode target = ((DumpReferenceToken) dumpToken).getValue();
        if (!target.getInvisible()) {
          result.put(dumpToken.getName(), target);
        }
      }
    }
    return result;
  }

  public static Map<String, Object> valueTokens(DumpNode node) {
    Map<String, Object> result = new HashMap<>();
    for (DumpToken dumpToken : node.getDumpTokenList()) {
      if (dumpToken.isDumpValueToken()) {
        // then it is a DumpValueToken
        DumpValueToken dumpValueToken = (DumpValueToken) dumpToken;
        result.put(dumpValueToken.getName(), dumpValueToken.getValue());
      }
    }
    return result;
  }

  public static List<DumpNode> invisiblePath(DumpNode node) {
    List<DumpNode> result = new ArrayList<>();
    for (InnerDumpNode inner : node.getInvisiblePath().getInnerDumpNodeList()) {
      DumpNode target = inner.getDumpNode();
      assertFalse(target.getInvisible());
      result.add(target);
    }
    return result;
  }

  static class ExposingDumpBuilder extends DumpBuilder {

    protected ExposingDumpBuilder(Object target) {
      super(target);
    }

    @Override
    public DumpAst build() {
      return super.build();
    }
  }

}
